from torch.nn.modules.loss import _Loss
from torch.nn import (
    MSELoss,
    L1Loss,
    SmoothL1Loss,
    HuberLoss,
    PoissonNLLLoss,
    KLDivLoss,
)


def get_loss_function(id: str) -> _Loss | None:
    match id.lower():
        case "mse":
            return MSELoss()
        case "l1":
            return L1Loss()
        case "smooth_l1":
            return SmoothL1Loss()
        case "huber":
            return HuberLoss()
        case "poisson":
            return PoissonNLLLoss()
        case "kl_div":
            return KLDivLoss()

    return None



from torch import Tensor
from torch.nn import Module, Linear, Sequential, ReLU, BatchNorm1d, Dropout


class FeedforwardRegressor(Module):
    def __init__(self, *,
                 input_dim: int,
                 hidden_dims: list[int],
                 dropout: float,
                 ) -> None:
        super().__init__()

        n_layers: int = len(hidden_dims)
        layers: list[Module] = []
        for i in range(n_layers):
            if i == 0:
                layers.append(Linear(input_dim, hidden_dims[i]))
            else:
                layers.append(Linear(hidden_dims[i-1], hidden_dims[i]))

            layers.append(BatchNorm1d(hidden_dims[i]))
            layers.append(ReLU())
            layers.append(Dropout(dropout))

        layers.append(Linear(hidden_dims[-1], 1))

        self.model: Sequential = Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.model(x)

        return x.squeeze()



import time

import numpy as np

import torch
from torch import Tensor
from torch.nn import Module
from torch.utils.data import DataLoader
from torch.optim import Optimizer
from torch.optim.lr_scheduler import StepLR

from torch.distributions import Distribution

from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.mlls import MarginalLogLikelihood
from gpytorch.distributions import MultivariateNormal

from vclog import Logger

from .regression_scaffold_interface import RegressionScaffold
from ...utils import get_optimizer, get_objective_function, ModelStateManager, log_train_data
from ...datasets import RegressionDataset
from ...dataclasses import RegressionEvalInfo, RegressionTrainInfo, RegressionPrediction


class GaussianRegressionScaffold(RegressionScaffold):
    def __init__(self, *,
                 model: Module,
                 dataset: RegressionDataset,
                 optimizer: str = "adam",
                 loss_function: str = "variational_elbo",
                 save_path: str | None = None,
                 ) -> None:
        self.device: torch.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.dtype: torch.dtype = torch.float32

        self.model: Module = model.to(self.device)
        self.likelihood = GaussianLikelihood().to(self.device)
        self.dataset: RegressionDataset = dataset

        self.optimizer_id: str = optimizer
        self.loss_function_id: str = loss_function

        if save_path is not None and "." in save_path:
            raise ValueError("save path must be a directory")
        self.save_path: str | None = save_path

        self.logger: Logger = Logger("catasta")

        # Logging info
        message: str = f"using {self.device} with {torch.cuda.get_device_name()}" if torch.cuda.is_available() else f"using {self.device}"
        self.logger.info(message)
        n_parameters: int = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        self.logger.info(f"training model {self.model.__class__.__name__} ({n_parameters} parameters)")

    def train(self,
              epochs: int = 100,
              batch_size: int = 32,
              lr: float = 1e-3,
              final_lr: float | None = None,
              early_stopping: tuple[int, float] | None = None,
              verbose: bool = True,
              ) -> RegressionTrainInfo:
        self.model.train()
        self.likelihood.train()

        optimizer: Optimizer | None = get_optimizer(self.optimizer_id, [self.model, self.likelihood], lr)
        if optimizer is None:
            raise ValueError(f"invalid optimizer id: {self.optimizer_id}")

        mll: MarginalLogLikelihood | None = get_objective_function(self.loss_function_id, self.model, self.likelihood, len(self.dataset.train))
        if mll is None:
            raise ValueError(f"invalid loss function id: {self.loss_function_id}")

        lr_decay: float = (final_lr / lr) ** (1 / epochs) if final_lr is not None else 1.0
        scheduler: StepLR = StepLR(optimizer, step_size=1, gamma=lr_decay)

        model_state_manager = ModelStateManager(early_stopping, self.save_path)

        data_loader: DataLoader = DataLoader(self.dataset.train, batch_size=batch_size, shuffle=False)

        time_per_epoch: float = 0.0

        losses: list[float] = []
        for i in range(epochs):
            batch_losses: list[float] = []
            start_time: float = time.time()
            for j, (x_batch, y_batch) in enumerate(data_loader):
                optimizer.zero_grad()

                x_batch = x_batch.to(self.device, dtype=self.dtype)
                y_batch = y_batch.to(self.device, dtype=self.dtype)

                output: MultivariateNormal = self.model(x_batch)

                loss: Tensor = -mll(output, y_batch)  # type: ignore
                loss.backward()

                optimizer.step()

                batch_losses.append(loss.item())

                if verbose:
                    log_train_data(
                        train_loss=losses[-1] if len(losses) > 0 else 0.0,
                        val_loss=0.0,
                        best_val_loss=0.0,
                        lr=scheduler.get_last_lr()[0],
                        epoch=i,
                        epochs=epochs,
                        percentage=int((i/epochs)*100+(j/len(data_loader))*100/epochs),
                        time_per_epoch=time_per_epoch,
                    )

            # END OF EPOCH
            scheduler.step()

            losses.append(np.mean(batch_losses).astype(float))

            model_state_manager(self.model.state_dict(), losses[-1])

            if model_state_manager.stop():
                self.logger.warning("early stopping")
                break

            time_per_epoch = time.time() - start_time

        # END OF TRAINING
        self.logger.info(f"training completed | best loss: {np.min(losses)}")

        model_state_manager.load_best_model_state(self.model)
        model_state_manager.save_models([self.model, self.likelihood])

        return RegressionTrainInfo(np.array(losses))

    @ torch.no_grad()
    def predict(self, input: np.ndarray | Tensor) -> RegressionPrediction:
        self.model.eval()
        self.likelihood.eval()

        input_tensor: Tensor = torch.tensor(input) if isinstance(input, np.ndarray) else input
        input_tensor = input_tensor.to(self.device, dtype=self.dtype)

        if input_tensor.ndim == 1:
            input_tensor = input_tensor.unsqueeze(0)

        output: Distribution = self.likelihood(self.model(input_tensor))

        mean: np.ndarray = output.mean.cpu().numpy()
        std: np.ndarray = output.stddev.cpu().numpy()

        return RegressionPrediction(mean, std)

    @torch.no_grad()
    def evaluate(self) -> RegressionEvalInfo:
        if self.dataset.test is None:
            raise ValueError("test split is empty")

        true_input: np.ndarray = np.array([])
        true_output: np.ndarray = np.array([])
        predicted_output: np.ndarray = np.array([])
        predicted_std: np.ndarray = np.array([])
        for x, y in self.dataset.test:
            true_input = np.append(true_input, x[-1])
            true_output = np.append(true_output, y)

            output: RegressionPrediction = self.predict(x)
            predicted_output = np.append(predicted_output, output.prediction)
            predicted_std = np.append(predicted_std, output.stds if output.stds is not None else np.zeros_like(output.prediction))

        return RegressionEvalInfo(true_input, true_output, predicted_output, predicted_std)



import numpy as np

from .transformation import Transformation


class Decimation(Transformation):
    def __init__(self, decimation_factor: int) -> None:
        self.decimation_factor: int = decimation_factor

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return x[::self.decimation_factor]



from vclog import Logger


def log_train_data(train_loss: float,
                   val_loss: float,
                   best_val_loss: float,
                   lr: float,
                   epoch: int,
                   epochs: int,
                   percentage: int,
                   time_per_epoch: float,
                   ) -> None:
    train_loss_str: str = f" {train_loss:.4f}" if train_loss > 0 else f"{train_loss:.4f}"
    val_loss_str: str = f"{val_loss:.4f}" if val_loss > 0 else f"{val_loss:.4f}"
    best_val_loss_str: str = f"{best_val_loss:.4f}" if best_val_loss > 0 else f"{best_val_loss:.4f}"
    lr_str: str = f"{lr:.2e}"
    epoch_str: str = f"{epoch}/{epochs}"
    percentage_str: str = f"{percentage}%" if percentage > 9.0 else f"0{percentage}%"

    time_remaining: int = int(time_per_epoch * (epochs - epoch))
    time_remaining_str: str = f"{int(time_remaining/60)}m {time_remaining%60}s remaining"

    log_str: str = f"epoch {epoch_str} ({percentage_str}) | loss[train, val, best] [{train_loss_str}, {val_loss_str}, {best_val_loss_str}] | lr: {lr_str} | {time_remaining_str}"

    Logger("catasta").info(log_str, flush=True)



from .regressors.approximate_gp_regressor import ApproximateGPRegressor
from .regressors.feedforward_regressor import FeedforwardRegressor
from .regressors.transformer_regressor.transformer_regressor import TransformerRegressor
from .regressors.rbf_regressor import RBFRegressor



from abc import ABC, abstractmethod
import numpy as np
from torch import Tensor
from ...dataclasses import RegressionEvalInfo, RegressionPrediction, RegressionTrainInfo


class RegressionScaffold(ABC):
    @abstractmethod
    def train(self, *,
              epochs: int,
              batch_size: int,
              lr: float,
              final_lr: float | None = None,
              early_stopping: tuple[int, float] | None = None,
              verbose: bool = True,
              ) -> RegressionTrainInfo:
        """
        Train the model

        Arguments
        ---------
        epochs: int
            Number of epochs to train for
        batch_size: int
            Batch size
        lr: float
            Learning rate
        final_lr: float
            Final learning rate for the learning rate scheduler. If None, then the learning rate will not decay.
        early_stopping: tuple[int, float]
            A tuple of (patience, min_delta) for the early stopping scheduler. If None, then the early stopping scheduler will not be used.
        verbose: bool
            Whether to print training logs

        Returns
        -------
        RegressionTrainInfo
            Training information
        """

    @abstractmethod
    def predict(self, input: np.ndarray | Tensor) -> RegressionPrediction:
        """
        Make predictions

        Arguments
        ---------
        input: np.ndarray | Tensor
            Input data

        Returns
        -------
        RegressionPrediction
            Prediction information
        """

    @abstractmethod
    def evaluate(self) -> RegressionEvalInfo:
        """
        Evaluate the model

        Returns
        -------
        RegressionEvalInfo
            Evaluation information
        """



from torch.nn import Module

from gpytorch.mlls import (
    MarginalLogLikelihood,
    VariationalELBO,
    PredictiveLogLikelihood,
    VariationalMarginalLogLikelihood,
    GammaRobustVariationalELBO,
)


def get_objective_function(id: str, model: Module, likelihood: Module, num_data: int) -> MarginalLogLikelihood | None:
    match id.lower():
        case "variational_elbo":
            return VariationalELBO(likelihood, model, num_data=num_data)
        case "predictive_log":
            return PredictiveLogLikelihood(likelihood, model, num_data=num_data)
        case "variational_marginal_log":
            return VariationalMarginalLogLikelihood(likelihood, model, num_data=num_data)
        case "gamma_robust_variational_elbo":
            return GammaRobustVariationalELBO(likelihood, model, num_data=num_data)

    return None



import numpy as np

from catasta.models import ApproximateGPRegressor
from catasta.datasets import RegressionDataset
from catasta.scaffolds import RegressionScaffold
from catasta.dataclasses import RegressionEvalInfo, RegressionTrainInfo
from catasta.transformations import (
    Normalization,
    Decimation,
    WindowSliding,
    Slicing,
)


def main() -> None:
    # Hyperparameters
    n_dim: int = 16
    n_inducing_points: int = 128
    kernel: str = "rq"
    mean: str = "constant"

    optimizer: str = "adamw"
    loss_function: str = "variational_elbo"

    epochs: int = 300
    batch_size: int = 128
    lr: float = 1e-3
    final_lr: float = 1e-4
    early_stopping: tuple[int, float] = (10, 1e-4)

    # Dataset
    dataset_root: str = "examples/data/"
    input_trasnsformations = [
        Slicing(amount=1, end="right"),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        WindowSliding(window_size=n_dim, stride=1),
    ]
    output_trasnsformations = [
        Slicing(amount=1, end="right"),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        Slicing(amount=n_dim - 1, end="left"),
    ]
    dataset = RegressionDataset(
        root=dataset_root,
        input_transformations=input_trasnsformations,
        output_transformations=output_trasnsformations,
        splits=(6/7, 0.0, 1/7),
    )

    # Model
    model = ApproximateGPRegressor(
        n_inducing_points=n_inducing_points,
        n_inputs=n_dim,
        kernel=kernel,
        mean=mean,
    )

    # Scaffold
    scaffold = RegressionScaffold(
        model=model,
        dataset=dataset,
        optimizer=optimizer,
        loss_function=loss_function,
    )

    train_info: RegressionTrainInfo = scaffold.train(
        epochs=epochs,
        batch_size=batch_size,
        lr=lr,
        final_lr=final_lr,
        early_stopping=early_stopping,
    )
    print(f"min train loss: {np.min(train_info.train_loss):.4f}")

    info: RegressionEvalInfo = scaffold.evaluate()
    print(info)


if __name__ == '__main__':
    main()



from typing import Callable

import torch
from torch import Tensor
from torch.nn import Module, Parameter, init


BasisFunction = Callable[[Tensor], Tensor]


def get_basis_function(id: str) -> BasisFunction | None:
    match id.lower():
        case "gaussian":
            return gaussian
        case "linear":
            return linear
        case "quadratic":
            return quadratic
        case "inverse_quadratic":
            return inverse_quadratic
        case "multiquadric":
            return multiquadric
        case "inverse_multiquadric":
            return inverse_multiquadric
        case "spline":
            return spline
        case "poisson_one":
            return poisson_one
        case "poisson_two":
            return poisson_two
        case "matern32":
            return matern32
        case "matern52":
            return matern52

    return None


def gaussian(alpha):
    phi = torch.exp(-1*alpha.pow(2))
    return phi


def linear(alpha):
    phi = alpha
    return phi


def quadratic(alpha):
    phi = alpha.pow(2)
    return phi


def inverse_quadratic(alpha):
    phi = torch.ones_like(alpha) / (torch.ones_like(alpha) + alpha.pow(2))
    return phi


def multiquadric(alpha):
    phi = (torch.ones_like(alpha) + alpha.pow(2)).pow(0.5)
    return phi


def inverse_multiquadric(alpha):
    phi = torch.ones_like(alpha) / (torch.ones_like(alpha) + alpha.pow(2)).pow(0.5)
    return phi


def spline(alpha):
    phi = (alpha.pow(2) * torch.log(alpha + torch.ones_like(alpha)))
    return phi


def poisson_one(alpha):
    phi = (alpha - torch.ones_like(alpha)) * torch.exp(-alpha)
    return phi


def poisson_two(alpha):
    phi = ((alpha - 2*torch.ones_like(alpha)) / 2*torch.ones_like(alpha)) \
        * alpha * torch.exp(-alpha)
    return phi


def matern32(alpha):
    phi = (torch.ones_like(alpha) + 3**0.5*alpha)*torch.exp(-3**0.5*alpha)
    return phi


def matern52(alpha):
    phi = (torch.ones_like(alpha) + 5**0.5*alpha + (5/3)
           * alpha.pow(2))*torch.exp(-5**0.5*alpha)
    return phi


class RBFRegressor(Module):
    def __init__(self, *,
                 in_features: int,
                 out_features: int,
                 basis_func: str = "gaussian",
                 ) -> None:
        super(RBFRegressor, self).__init__()

        self.in_features: int = in_features
        self.out_features: int = out_features

        self.centres = Parameter(Tensor(out_features, in_features))
        self.log_sigmas = Parameter(Tensor(out_features))

        basis_function: BasisFunction | None = get_basis_function(basis_func)
        if basis_function is None:
            raise ValueError(f"Unknown basis function: {basis_func}")

        self.basis_function = basis_function

        self.reset_parameters()

    def reset_parameters(self) -> None:
        init.normal_(self.centres, 0, 1)
        init.constant_(self.log_sigmas, 0)

    def forward(self, input: Tensor) -> Tensor:
        size = (input.size(0), self.out_features, self.in_features)
        x = input.unsqueeze(1).expand(size)
        c = self.centres.unsqueeze(0).expand(size)
        distances = (x - c).pow(2).sum(-1).pow(0.5) / torch.exp(self.log_sigmas).unsqueeze(0)

        return self.basis_function(distances).squeeze(-1)



from typing import Callable

import numpy as np

from .transformation import Transformation


class Custom(Transformation):
    def __init__(self, function: Callable) -> None:
        """
        Initialize the custom transformation

        Parameters
        ----------
        function : Callable
            Custom transformation function
        """
        self.function: Callable = function

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Apply custom transformation to the input data

        Parameters
        ----------
        x : np.ndarray
            Input data to be transformed

        Returns
        -------
        np.ndarray
            Transformed data
        """
        return self.function(x)



import numpy as np

from .transformation import Transformation


class WindowSliding(Transformation):
    def __init__(self, window_size: int, stride: int) -> None:
        self.window_size: int = window_size
        self.stride: int = stride

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return np.array(
            [
                x[i: i + self.window_size]
                for i in range(0, len(x) - self.window_size + 1, self.stride)
            ]
        )



import torch
from torch import Tensor
from torch.nn import (
    Module,
    Linear,
    LayerNorm,
    Softmax,
    Sequential,
    Dropout,
)

from einops import rearrange


class Attention(Module):
    def __init__(self, *,
                 d_model: int,
                 n_heads: int,
                 head_dim: int,
                 dropout: float,
                 ) -> None:
        super().__init__()

        self.n_heads: int = n_heads
        self.scale: int = head_dim ** -0.5
        self.norm = LayerNorm(d_model)
        self.attend = Softmax(dim=-1)
        self.dropout = Dropout(dropout)

        inner_dim: int = head_dim * n_heads
        self.to_qkv = Linear(d_model, inner_dim * 3, bias=False)
        self.to_out = Sequential(
            Linear(inner_dim, d_model, bias=False),
            Dropout(dropout),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.norm(x)

        qkv: Tensor = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.n_heads), qkv)

        dots: Tensor = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn: Tensor = self.attend(dots)
        attn = self.dropout(attn)

        out: Tensor = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')

        return self.to_out(out)



from .optimizer_factory import get_optimizer
from .loss_function_factory import get_loss_function
from .objective_function_factory import get_objective_function
from .training_logger import log_train_data
from .model_state_manager import ModelStateManager



from torch.nn import Module
from gpytorch.models.gp import GP

from .regression_scaffold_interface import RegressionScaffold as IRegressionScaffold
from ...datasets import RegressionDataset
from .vanilla_regression_scaffold import VanillaRegressionScaffold
from .gaussian_regression_scaffold import GaussianRegressionScaffold


def RegressionScaffold(*,
                       model: Module,
                       dataset: RegressionDataset,
                       optimizer: str,
                       loss_function: str,
                       save_path: str | None = None,
                       ) -> IRegressionScaffold:
    '''
    Create a scaffold for regression tasks.

    Arguments
    ---------
    model: Module
        The model to train. It must be a Catasta regressor.

    dataset: RegressionDataset
        The dataset to train on.

    optimizer: str
        The ID of the optimizer to use. Available optimizers are:
            - "adam"
            - "adamw"
            - "sgd"
            - "lbfgs"
            - "rmsprop"
            - "rprop"
            - "adadelta"
            - "adagrad"
            - "adamax"
            - "asgd"
            - "sparse_adam"

    loss_function: str
        The ID of the loss function to use. Available loss functions are:
            - "mse"
            - "l1"
            - "smooth_l1"
            - "huber"
            - "poisson" 
            - "kl_div"
            - "variational_elbo" (only for Gaussian processes)
            - "predictive_log" (only for Gaussian processes)
            - "variational_marginal_log" (only for Gaussian processes)

    save_path: str | None
        The path where to save the model. If None, the model is not saved. It must be a directory.

    Returns
    -------
    RegressionScaffold
        The scaffold class to train the model.
    '''
    match model:
        case GP():
            return GaussianRegressionScaffold(
                model=model,
                dataset=dataset,
                optimizer=optimizer,
                loss_function=loss_function,
                save_path=save_path,
            )
        case _:
            return VanillaRegressionScaffold(
                model=model,
                dataset=dataset,
                optimizer=optimizer,
                loss_function=loss_function,
                save_path=save_path,
            )



from torch import Tensor
from torch.nn import (
    Module,
    Linear,
    LayerNorm,
    Sequential,
    GELU,
    Dropout,
)


class FeedForward(Module):
    def __init__(self, *,
                 d_model: int,
                 hidden_dim: int,
                 dropout: float,
                 ) -> None:
        super().__init__()

        self.net = Sequential(
            LayerNorm(d_model),
            Linear(d_model, hidden_dim),
            GELU(),
            Dropout(dropout),
            Linear(hidden_dim, d_model),
            Dropout(dropout),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.net(x)



import torch
from torch import Tensor


def posemb_sincos_1d(patches: Tensor, temperature: int = 10000) -> Tensor:
    n: int = patches.shape[1]
    d_model: int = patches.shape[2]
    device: torch.device = patches.device
    dtype: torch.dtype = patches.dtype

    if (d_model % 2) != 0:
        raise ValueError(f'feature dimension must be multiple of 2 for sincos emb. got {d_model}')

    n_tensor: Tensor = torch.arange(n, device=device)
    omega: Tensor = torch.arange(d_model // 2, device=device) / (d_model // 2 - 1)
    omega = 1.0 / (temperature ** omega)

    n_tensor = n_tensor.flatten()[:, None] * omega[None, :]
    pe: Tensor = torch.cat((n_tensor.sin(), n_tensor.cos()), dim=1)

    return pe.to(dtype)



import os

import torch
from torch.nn import Module

from vclog import Logger


class ModelStateManager:
    def __init__(self, patience_and_delta: tuple[int, float] | None, save_path: str | None) -> None:
        self.patience: int | None = patience_and_delta[0] if patience_and_delta is not None else None
        self.delta: float | None = patience_and_delta[1] if patience_and_delta is not None else None

        self.save_path: str | None = save_path
        if self.save_path is not None:
            if not os.path.exists(self.save_path):
                os.makedirs(self.save_path)

        self.best_loss: float = float("inf")
        self.counter: int = 0
        self.best_model_state: dict = {}
        self.early_stop: bool = False

        self.logger: Logger = Logger("catasta")

    def __call__(self, model_state: dict, loss: float):
        if loss < self.best_loss:
            self.best_loss = loss
            self.counter = 0
            self.best_model_state = model_state
        elif self.delta is not None and self.patience is not None:
            if loss - self.best_loss > self.delta:
                self.counter += 1

        if self.patience is not None and self.delta is not None:
            if self.counter >= self.patience:
                self.early_stop = True

        if not self.best_model_state:
            self.best_model_state = model_state

    def stop(self) -> bool:
        return self.early_stop

    def save_models(self, models: list[Module]) -> None:
        if self.save_path is None:
            return

        for model in models:
            model_name: str = model.__class__.__name__
            model_path: str = os.path.join(self.save_path, f"{model_name}.pt")
            torch.save(model.state_dict(), model_path)

            self.logger.info(f"saved model {model_name} to {model_path}")

    def load_best_model_state(self, model: Module) -> None:
        model.load_state_dict(self.best_model_state)



import torch
from torch import Tensor

from gpytorch.models import ApproximateGP
from gpytorch.variational import CholeskyVariationalDistribution, VariationalStrategy
from gpytorch.means import ZeroMean, ConstantMean, Mean
from gpytorch.kernels import Kernel, ScaleKernel, RBFKernel, MaternKernel, RQKernel, RFFKernel, PeriodicKernel
from gpytorch.distributions import MultivariateNormal


def _get_kernel(id: str, n_inputs: int) -> Kernel | None:
    match id.lower():
        case "rq":
            return RQKernel(ard_num_dims=n_inputs)
        case "matern":
            return MaternKernel(ard_num_dims=n_inputs)
        case "rbf":
            return RBFKernel(ard_num_dims=n_inputs)
        case "rff":
            return RFFKernel(num_samples=n_inputs)
        case "periodic":
            return PeriodicKernel(ard_num_dims=n_inputs)

    return None


def _get_mean_module(id: str) -> Mean | None:
    match id.lower():
        case "constant":
            return ConstantMean()
        case "zero":
            return ZeroMean()

    return None


class ApproximateGPRegressor(ApproximateGP):
    def __init__(self, *,
                 n_inducing_points: int,
                 n_inputs: int,
                 kernel: str = "rq",
                 mean: str = "constant"
                 ) -> None:
        '''
        Arguments
        ---------
        n_inducing_points: int
            Number of inducing points
        n_inputs: int
            Number of input dimensions
        kernel: str
            Kernel to use. One of "rq", "matern", "rbf", "rff", "periodic"
        mean: str
            Mean to use. One of "constant", "zero"
        '''

        self.n_inducing_points: int = n_inducing_points
        self.n_inputs: int = n_inputs
        dtype: torch.dtype = torch.float32

        inducing_points: torch.Tensor = torch.randn(n_inducing_points, n_inputs, dtype=dtype)

        variational_distribution = CholeskyVariationalDistribution(n_inducing_points)
        variational_strategy = VariationalStrategy(self, inducing_points, variational_distribution, learn_inducing_locations=True)

        super().__init__(variational_strategy)

        mean_module: Mean | None = _get_mean_module(mean)
        if mean_module is None:
            raise ValueError(f"Unknown mean: {mean}")

        self.mean_module = mean_module

        kernel_module: Kernel | None = _get_kernel(kernel, n_inputs)
        if kernel_module is None:
            raise ValueError(f"Unknown kernel: {kernel}")

        self.covar_module = ScaleKernel(kernel_module)

    def forward(self, x: Tensor) -> MultivariateNormal:
        mean_x: Tensor = self.mean_module(x)  # type: ignore
        covar_x: Tensor = self.covar_module(x)  # type: ignore

        return MultivariateNormal(mean_x, covar_x)



from torch.nn import Module
from gpytorch.models.gp import GP

from .regression_archway_interface import RegressionArchway as IRegressionArchway
from .vanilla_regression_archway import VanillaRegressionArchway
from .gaussian_regression_archway import GaussianRegressionArchway


def RegressionArchway(*,
                      model: Module,
                      path: str | None = None,
                      ) -> IRegressionArchway:
    '''
    Create an "archway" for regression inference.

    Arguments
    ---------
    model: Module
        The model to use for inference.

    path: str | None
        The path to the pretrained model. If None, the model is not loaded.

    Returns
    -------
    RegressionArchway
        The archway class to perform inference.
    '''

    match model:
        case GP():
            return GaussianRegressionArchway(
                model=model,
                path=path
            )
        case _:
            return VanillaRegressionArchway(
                model=model,
                path=path
            )



import time

import numpy as np

import torch
from torch import Tensor
from torch.nn import Module
from torch.utils.data import DataLoader
from torch.optim import Optimizer
from torch.optim.lr_scheduler import StepLR
from torch.nn.modules.loss import _Loss

from vclog import Logger

from .regression_scaffold_interface import RegressionScaffold
from ...datasets import RegressionDataset
from ...dataclasses import RegressionEvalInfo, RegressionPrediction, RegressionTrainInfo
from ...utils import get_optimizer, get_loss_function, log_train_data, ModelStateManager


class VanillaRegressionScaffold(RegressionScaffold):
    def __init__(self, *,
                 model: Module,
                 dataset: RegressionDataset,
                 optimizer: str = "adam",
                 loss_function: str = "mse",
                 save_path: str | None = None,
                 ) -> None:
        self.device: torch.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.dtype: torch.dtype = torch.float32

        self.model: Module = model.to(self.device)

        self.dataset: RegressionDataset = dataset

        self.optimmizer_id: str = optimizer
        self.loss_function_id: str = loss_function

        if save_path is not None and "." in save_path:
            raise ValueError("save path must be a directory")
        self.save_path: str | None = save_path

        self.logger: Logger = Logger("catasta")

        # Logging info
        message: str = f"using {self.device} with {torch.cuda.get_device_name()}" if torch.cuda.is_available() else f"using {self.device}"
        self.logger.info(message)
        n_parameters: int = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        self.logger.info(f"training model {self.model.__class__.__name__} ({n_parameters} parameters)")

    def train(self, *,
              epochs: int = 100,
              batch_size: int = 128,
              lr: float = 1e-3,
              final_lr: float | None = None,
              early_stopping: tuple[int, float] | None = None,
              verbose: bool = True,
              ) -> RegressionTrainInfo:
        self.model.train()

        if self.dataset.validation is None:
            self.logger.warning("no validation split found")

        optimizer: Optimizer | None = get_optimizer(self.optimmizer_id, self.model, lr)
        if optimizer is None:
            raise ValueError(f"invalid optimizer id: {self.optimmizer_id}")

        loss_function: _Loss | None = get_loss_function(self.loss_function_id)
        if loss_function is None:
            raise ValueError(f"invalid loss function id: {self.loss_function_id}")

        lr_decay: float = (final_lr / lr) ** (1 / epochs) if final_lr is not None else 1.0
        scheduler: StepLR = StepLR(optimizer, step_size=1, gamma=lr_decay)

        model_state_manager = ModelStateManager(early_stopping, self.save_path)

        data_loader: DataLoader = DataLoader(self.dataset.train, batch_size=batch_size, shuffle=True)

        eval_loss: float = np.inf

        time_per_epoch: float = 0.0

        train_losses: list[float] = []
        eval_losses: list[float] = []
        for i in range(epochs):
            batch_train_losses: list[float] = []
            start_time: float = time.time()
            for j, (x_batch, y_batch) in enumerate(data_loader):
                optimizer.zero_grad()

                x_batch = x_batch.to(self.device, dtype=self.dtype)
                y_batch = y_batch.to(self.device, dtype=self.dtype)

                output: Tensor = self.model(x_batch)

                loss: Tensor = loss_function(output, y_batch)
                loss.backward()

                optimizer.step()

                batch_train_losses.append(loss.item())

                if verbose:
                    log_train_data(
                        train_loss=train_losses[-1] if len(train_losses) > 0 else 0,
                        val_loss=eval_loss,
                        best_val_loss=model_state_manager.best_loss,
                        lr=scheduler.get_last_lr()[0],
                        epoch=i,
                        epochs=epochs,
                        percentage=int((i/epochs)*100+(j/len(data_loader))*100/epochs),
                        time_per_epoch=time_per_epoch,
                    )

            # END OF EPOCH
            scheduler.step()

            train_losses.append(np.mean(batch_train_losses).astype(float))

            eval_loss = self._estimate_loss(batch_size)
            eval_losses.append(eval_loss)

            model_state_manager(self.model.state_dict(), eval_loss)

            if model_state_manager.stop():
                self.logger.warning("early stopping")
                break

            time_per_epoch = time.time() - start_time

        # END OF TRAINING
        self.logger.info(f'training completed | best eval loss: {np.min(eval_losses):.4f}')

        model_state_manager.load_best_model_state(self.model)
        model_state_manager.save_models([self.model])

        return RegressionTrainInfo(np.array(train_losses), np.array(eval_losses))

    @torch.no_grad()
    def predict(self, input: np.ndarray | Tensor) -> RegressionPrediction:
        self.model.eval()

        input_tensor: Tensor = torch.tensor(input) if isinstance(input, np.ndarray) else input
        input_tensor = input_tensor.to(self.device, dtype=self.dtype)

        if len(input_tensor.shape) == 1:
            input_tensor = input_tensor.unsqueeze(0)

        output: Tensor = self.model(input_tensor)

        return RegressionPrediction(output.cpu().numpy())

    @torch.no_grad()
    def evaluate(self) -> RegressionEvalInfo:
        if self.dataset.test is None and self.dataset.validation is not None:
            self.logger.warning("no test split found, using validation split for evaluation")
            self.dataset.test = self.dataset.validation
        else:
            raise ValueError(f"cannot evaluate without a test split")

        true_input: np.ndarray = np.array([])
        true_output: np.ndarray = np.array([])
        predicted_output: np.ndarray = np.array([])
        for x, y in self.dataset.test:
            output: RegressionPrediction = self.predict(x)

            true_input = np.append(true_input, x[-1])
            true_output = np.append(true_output, y)
            predicted_output = np.append(predicted_output, output.prediction)

        return RegressionEvalInfo(true_input, true_output, predicted_output)

    @torch.no_grad()
    def _estimate_loss(self, batch_size: int) -> float:
        if self.dataset.validation is None:
            return np.inf

        self.model.eval()

        loss_function: _Loss | None = get_loss_function(self.loss_function_id)
        if loss_function is None:
            raise ValueError(f"invalid loss function id: {self.loss_function_id}")

        data_loader: DataLoader = DataLoader(self.dataset.validation, batch_size=batch_size, shuffle=False)

        losses: list[float] = []
        for x_batch, y_batch in data_loader:
            x_batch: Tensor = x_batch.to(self.device, dtype=self.dtype)
            y_batch: Tensor = y_batch.to(self.device, dtype=self.dtype)

            output: Tensor = self.model(x_batch)

            loss: Tensor = loss_function(output, y_batch)

            losses.append(loss.item())

        self.model.train()

        return np.mean(losses).astype(float)



import torch
from torch import Tensor
from torch.nn import (
    Module,
    Linear,
    LayerNorm,
    Sequential,
    ModuleList,
    Dropout,
)
from torch.fft import fft

from einops import rearrange
from einops.layers.torch import Rearrange

from .modules import (
    posemb_sincos_1d,
    Attention,
    FeedForward,
)


class Transformer(Module):
    def __init__(self, *,
                 d_model: int,
                 n_layers: int,
                 n_heads: int,
                 head_dim: int,
                 feedforward_dim: int,
                 dropout: float,
                 ) -> None:
        super().__init__()

        self.norm = LayerNorm(d_model)
        self.layers = ModuleList([])
        for _ in range(n_layers):
            self.layers.append(ModuleList([
                Attention(d_model=d_model,
                          n_heads=n_heads,
                          head_dim=head_dim,
                          dropout=dropout,
                          ),
                FeedForward(d_model=d_model,
                            hidden_dim=feedforward_dim,
                            dropout=dropout,
                            )
            ]))

    def forward(self, x: Tensor) -> Tensor:
        for attn, ff in self.layers:  # type: ignore
            x = attn(x) + x
            x = ff(x) + x

        return self.norm(x)


class TransformerRegressor(Module):
    def __init__(self, *,
                 context_length: int,
                 n_patches: int,
                 d_model: int,
                 n_layers: int,
                 n_heads: int,
                 feedforward_dim: int,
                 head_dim: int,
                 dropout: float,
                 use_fft: bool = False,
                 ) -> None:
        super().__init__()
        patch_size: int = context_length // n_patches
        freq_patch_dim: int = patch_size * 2

        if context_length % patch_size != 0:
            raise ValueError(f"sequence length {context_length} must be divisible by patch size {patch_size}")

        self.use_fft = use_fft

        self.to_patch_embedding = Sequential(
            Rearrange('b c (n p) -> b n (p c)', p=patch_size),
            LayerNorm(patch_size),
            Linear(patch_size, d_model),
            LayerNorm(d_model),
        )

        self.to_freq_embedding = Sequential(
            Rearrange('b c (n p) ri -> b n (p ri c)', p=patch_size),
            LayerNorm(freq_patch_dim),
            Linear(freq_patch_dim, d_model),
            LayerNorm(d_model),
        )

        self.pos_embedding = posemb_sincos_1d
        self.dropout = Dropout(dropout)

        self.transformer = Transformer(
            d_model=d_model,
            n_layers=n_layers,
            n_heads=n_heads,
            head_dim=head_dim,
            feedforward_dim=feedforward_dim,
            dropout=dropout,
        )

        self.linear_head = Sequential(
            LayerNorm(d_model),
            Linear(d_model, 1),
        )

    def forward(self, x: Tensor) -> Tensor:
        if self.use_fft:
            return self.fft_forward(x)
        else:
            return self.no_fft_forward(x)

    def no_fft_forward(self, input: Tensor) -> Tensor:
        input = rearrange(input, 'b s -> b 1 s')

        x = self.to_patch_embedding(input)
        x += self.pos_embedding(x)

        x = self.transformer(x)
        x = x.mean(dim=1)

        x = self.linear_head(x).squeeze()

        return x

    def fft_forward(self, input: Tensor) -> Tensor:
        input = rearrange(input, 'b s -> b 1 s')
        freqs: Tensor = torch.view_as_real(fft(input))

        x = self.to_patch_embedding(input)
        f = self.to_freq_embedding(freqs)

        x += self.pos_embedding(x)
        f += self.pos_embedding(f)

        x = torch.cat((x, f), dim=1)

        x = self.transformer(x)
        x = x.mean(dim=1)

        x = self.linear_head(x).squeeze()

        return x



import os
import pandas as pd
import numpy as np

import torch
from torch import Tensor
from torch.utils.data import Dataset, Subset

from ..transformations import Transformation


class RegressionDataset(Dataset):
    def __init__(self, *,
                 root: str,
                 input_transformations: list[Transformation] = [],
                 output_transformations: list[Transformation] = [],
                 splits: tuple[float, float, float] = (0.8, 0.1, 0.1),
                 ) -> None:
        self.root: str = root if root.endswith("/") else root + "/"
        self.train_split: float = splits[0]
        self.validation_split: float = splits[1]
        self.test_split: float = splits[2]

        self.input_transformations: list[Transformation] = input_transformations
        self.output_transformations: list[Transformation] = output_transformations

        self.inputs: np.ndarray = np.array([])
        self.outputs: np.ndarray = np.array([])

        self._check_arguments()
        self.inputs, self.outputs = self._prepare_data(*self._get_data())
        self.train, self.validation, self.test = self._split_dataset()

    def _check_arguments(self) -> None:
        if not os.path.isdir(self.root):
            raise ValueError(f"root must be a directory. Found {self.root}")
        splits_sum: float = round(sum([self.train_split, self.validation_split, self.test_split]), 4)
        if splits_sum != 1:
            raise ValueError(f"splits must sum to 1. Found {splits_sum}")
        if self.validation_split == 0 and self.test_split == 0:
            raise ValueError(f"at least a validation or test split must be greater than 0")
        if self.train_split == 0:
            raise ValueError(f"train split must be greater than 0")

    def _get_data(self) -> tuple[list[np.ndarray], list[np.ndarray]]:
        inputs: list[np.ndarray] = []
        outputs: list[np.ndarray] = []

        filename_counter: int = 0
        filenames: list[str] = os.listdir(self.root)
        filenames.sort()
        for filename in filenames:
            if not filename.endswith(".csv"):
                continue
            filename_counter += 1

            data_frame: pd.DataFrame = pd.read_csv(self.root + filename)

            if 'input' not in data_frame.columns or 'output' not in data_frame.columns:
                raise ValueError(f"CSV file {filename} must have the columns 'input' and 'output'")

            inputs.append(data_frame['input'].to_numpy().flatten())
            outputs.append(data_frame['output'].to_numpy().flatten())

            if len(inputs[-1]) != len(outputs[-1]):
                raise ValueError(f"CSV file {filename} must have the same number of rows for 'input' and 'output'")

        if filename_counter == 0:
            raise ValueError(f"Directory {self.root} must contain at least one CSV file")

        return inputs, outputs

    def _prepare_data(self, inputs: list[np.ndarray], outputs: list[np.ndarray]) -> tuple[np.ndarray, np.ndarray]:
        transformed_inputs: list[np.ndarray] = []
        transformed_outputs: list[np.ndarray] = []

        for input, output in zip(inputs, outputs):
            for transformation in self.input_transformations:
                input = transformation(input)
            for transformation in self.output_transformations:
                output = transformation(output)

            transformed_inputs.append(input)
            transformed_outputs.append(output)

        inputs_array: np.ndarray = np.concatenate(transformed_inputs)
        outputs_array: np.ndarray = np.concatenate(transformed_outputs)

        return inputs_array, outputs_array

    def _split_dataset(self) -> tuple[Subset, Subset | None, Subset | None]:
        train_index: int = int(self.train_split * len(self))
        validation_index: int = train_index + int(self.validation_split * len(self))
        test_index: int = validation_index + int(self.test_split * len(self))
        train: Subset = Subset(self, range(train_index))
        validation: Subset | None = Subset(self, range(train_index, validation_index)) if self.validation_split > 0 else None
        test: Subset | None = Subset(self, range(validation_index, test_index)) if self.test_split > 0 else None

        return train, validation, test

    def __len__(self) -> int:
        return self.inputs.shape[0]

    def __getitem__(self, index: int) -> tuple[Tensor, Tensor]:
        return torch.tensor(self.inputs[index]), torch.tensor(self.outputs[index]).squeeze()



import os

import numpy as np

import torch
from torch import Tensor
from torch.nn import Module
from torch.distributions import Distribution

from gpytorch.likelihoods import GaussianLikelihood

from vclog import Logger

from .regression_archway_interface import RegressionArchway
from ...dataclasses import RegressionPrediction


class GaussianRegressionArchway(RegressionArchway):
    def __init__(self, *,
                 model: Module,
                 path: str | None = None,
                 ) -> None:
        self.device: torch.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.dtype: torch.dtype = torch.float32

        self.logger = Logger("catasta")

        self.model: Module = model.to(self.device, dtype=self.dtype)
        self.likelihood = GaussianLikelihood().to(self.device)
        self.load_models(path)

        self.model.eval()
        self.likelihood.eval()

        # Logging info
        message: str = f"using {self.device} with {torch.cuda.get_device_name()}" if torch.cuda.is_available() else f"using {self.device}"
        self.logger.info(message)
        n_parameters: int = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        self.logger.info(f"infering with {self.model.__class__.__name__} ({n_parameters} parameters)")

    def load_models(self, path: str | None) -> None:
        if path is None:
            return

        # get the .pt files in the directory
        files: list[str] = os.listdir(path)

        for file in files:
            if not file.endswith(".pt"):
                continue

            if self.model.__class__.__name__ in file:
                self.logger.info(f"Loading model from {file}")
                self.model.load_state_dict(torch.load(os.path.join(path, file)))
            elif self.likelihood.__class__.__name__ in file:
                self.logger.info(f"Loading likelihood from {file}")
                self.likelihood.load_state_dict(torch.load(os.path.join(path, file)))

    @torch.no_grad()
    def predict(self, input: np.ndarray | Tensor) -> RegressionPrediction:
        input_tensor: Tensor = torch.tensor(input) if isinstance(input, np.ndarray) else input
        input_tensor = input_tensor.to(self.device, dtype=self.dtype)

        if input_tensor.ndim == 1:
            input_tensor = input_tensor.unsqueeze(0)

        output: Distribution = self.likelihood(self.model(input_tensor))

        mean: np.ndarray = output.mean.cpu().numpy()
        std: np.ndarray = output.stddev.cpu().numpy()

        return RegressionPrediction(mean, std)

    def __call__(self, input: np.ndarray | Tensor) -> RegressionPrediction:
        return self.predict(input)



from abc import ABC, abstractmethod
import numpy as np
from torch import Tensor
from ...dataclasses import RegressionPrediction


class RegressionArchway(ABC):
    @abstractmethod
    def predict(self, input: np.ndarray | Tensor) -> RegressionPrediction:
        """
        Make predictions

        Arguments
        ---------
        input: np.ndarray | Tensor
            Input data

        Returns
        -------
        RegressionPrediction
            Prediction information
        """



import numpy as np

from .transformation import Transformation
from .window_sliding import WindowSliding


class DiffAndConcat(Transformation):
    def __init__(self, *,
                 n_diffs: int,
                 elements_per_diff: int,
                 filter: Transformation,
                 ) -> None:
        self.n_diffs: int = n_diffs
        self.elements_per_diff: int = elements_per_diff
        self.filter: Transformation = filter
        self.window_sliding: WindowSliding = WindowSliding(
            window_size=elements_per_diff,
            stride=1,
        )

    def __call__(self, x: np.ndarray) -> np.ndarray:
        diffs: np.ndarray = np.zeros((self.n_diffs+1, len(x)))
        diffs[0] = x
        for i in range(1, self.n_diffs+1):
            diffs[i] = np.gradient(self.filter(diffs[i - 1]))

        x = np.concatenate([self.window_sliding(diff) for diff in diffs], axis=1)

        return x.flatten()
